package org.greenstone.gatherer.gui;
/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;
import javax.swing.tree.*;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.file.FileNode;
import org.greenstone.gatherer.file.FileOpenActionListener;
import org.greenstone.gatherer.gui.Filter;
import org.greenstone.gatherer.gui.GComboBox;
import org.greenstone.gatherer.gui.MetaEditPrompt;
import org.greenstone.gatherer.gui.WarningDialog;
import org.greenstone.gatherer.gui.table.GTableModel;
import org.greenstone.gatherer.gui.table.TableCellRenderer;
import org.greenstone.gatherer.gui.tree.DragTree;
import org.greenstone.gatherer.msm.ElementWrapper;
import org.greenstone.gatherer.msm.Metadata;
import org.greenstone.gatherer.msm.MSMEvent;
import org.greenstone.gatherer.msm.MSMListener;
import org.greenstone.gatherer.util.ArrayTools;
import org.greenstone.gatherer.util.DragGroup;
import org.greenstone.gatherer.util.TreeSynchronizer;
import org.greenstone.gatherer.util.Utility;
import org.greenstone.gatherer.util.XORToggleButtonGroup;
import org.greenstone.gatherer.valuetree.GValueTree;
import org.greenstone.gatherer.valuetree.GValueModel;
import org.greenstone.gatherer.valuetree.GValueNode;
import org.w3c.dom.Element;
/** Provides a view of controls for the editing of metadata. It makes use of several other important components such as a GTree, GTable and GValueTree. While much of the gui content is left to these components, the MetaEditPane is resposible for actioning metadata edit requests, listing for mouse clicks within its scope and other data functionality (such as providing a list of the selected files).
* @author John Thompson, Greenstone Digital Libraries, University of Waikato
* @version 2.3b
*/
public class MetaEditPane
    extends JPanel
    implements ActionListener, ListSelectionListener, TreeModelListener, TreeSelectionListener, MSMListener {
    /** <i>true</i> if the selection has changed since the last time it was asked for, <i>false</i> otherwise. */
    public boolean has_changed = false;
    /** The GValueTree graphically shows the available metadata that has been previously assigned, and provides controls for adding, updating and removing metadata, depending on the users selections in the other important components. */
    public GValueTree tree = null;
    /** If only files are selected, then show folder is not available. */
    private boolean show_folder_enabled = false;
    /** The layout manager used to display either the GTable (when records are selected) or a placeholder panel with the immortal words 'No Record Selected'. */
    private CardLayout card_layout = null;
    /** The layout manager used to display either the GValueTree (when records are selected) or a placeholder panel with the immortal words 'No Record Selected'. */
    private CardLayout card_layout2 = null;
    /** Used to dynamically filter the collection tree at user request. Note that this is synchronized with the collection tree filter in the Gather view. */
    private Filter filter = null;
    /** The data model behind the GTable, to which we hold a reference for convenient access. */
    private GTableModel model = null;
    /** The currently reported selection. Note that this may actually be incorrect, but if so the changed flag will be set, and the selection will be updated the next time it is needed. */
    private FileNode records[] = null;
    /** Records which of the three GTable views is currently being shown, so we can determine what view to change to next (as they cycle). */
    //private int current_table_view;
    /** The mode is determined by the metadata edit action choosen. */
    private int mode;
    /** The button, which when clicked, adds metadata to the selected records. */
    private JButton add;
    /** The button, which when clicked, removes the selected metadata from the selected records. */
    private JButton remove;
    /** The button, which when clicked, changes the currently shown GTable view. */
    //private JButton table_view;
    /** The button, which when clicked, updates the selected metadata from the selected records. */
    private JButton update;
    /** This button creates an editor dialog to allow for an expanded area to type in text. */
    private JButton expand;
    /** The label at the top of the collection tree, which shows the collection name. */
    private JLabel collection_label;
    /** The panel in which the metadata value card layout resides. */
    private JPanel control_pane;
    /** The panel in which the metadata table card layout resides. */
    private JPanel table_card_pane;
    /** The panel the table is added to. */
    private JPanel table_pane;
    /** An icon showing the current state of the table build. */
    //private JProgressBar activity_bar;
    /** The splitpane dividing the collection tree and the metadata based controls. */
    private JSplitPane external_split;
    /** Divides the metadata table and the value tree controls. */
    private JSplitPane main_split_pane;
    /** A reference to the metadata table, via its superclass. */
    private JTable table;
    /** The label shown at the top of the metadata table detailing the current selection statistics. */
    private JTextField table_label;
    /** The button to control whether assigned metadata is shown. */
    //private JToggleButton assigned_metadata_view;
    /** The button to control whether unassigned metadata is shown. */
    //private JToggleButton unassigned_metadata_view;
    /** A reference to the collection tree. */
    public DragTree collection_tree;
    /** The currently selected metadata determined by listening to every second list selection event from the metadata table. */
    private Metadata selected_metadata;
    /** Listens for right clicks over the collection tree. */
    private RightButtonListener right_button_listener = null;
    /** A temporary storage array from Strings passed to the dictionary to be inserted in the phrase returned. */
    private String args[];
    /** Provide synchronization between the collection trees in this view and the collection pane view. */
    private TreeSynchronizer tree_sync = null;
    static public Dimension BUTTON_SIZE = new Dimension(190, 25);
    static private Dimension CONTROL_SIZE = new Dimension(560, 240);
    static private Dimension LABEL_SIZE = new Dimension(75, 25);
    static private Dimension MINIMUM_SIZE = new Dimension(100, 100);
    static private Dimension TABLE_SIZE = new Dimension(560, 25);
    static private Dimension TREE_SIZE = new Dimension(250, 500);
    /** An element of the Mode type enumeration, indicating we are adding metadata (ie no metadata rows are selected in the metadata table). */
    static private int ADD = 0;
    /** An element of the Mode type enumeration, indicating we are editing existing metadata (ie some row is selected in the metadata table). */
    static private int EDIT = 1;
    /** The name of the panel containing the metadata table. */
    static final private String CARD_ONE  = "Card One";
    /** The name of the panel containing the 'no file' placeholder for the metadata table. */
    static final private String CARD_TWO = "Card Two";
    /** The name of the panel containing the 'no metadata' placeholder for the metadata table. */
    static final private String CARD_ZERO = "Card Zero";
    /** The name of the panel containing the placeholder for the value tree. */
    static final private String TOOLS_OFF = "Tools Off";
    /** The name of the panel containing the value tree. */
    static final private String TOOLS_ON  = "Tools On";
    /** Constructor.
     * @param tree_sync The <strong>TreeSynchronizer</strong> to be used on the collection tree.
     * @see org.greenstone.gatherer.Configuration
     * @see org.greenstone.gatherer.gui.table.GTable
     * @see org.greenstone.gatherer.valuetree.GValueTree
     */
    public MetaEditPane(TreeSynchronizer tree_sync) {
	//this.current_table_view = GTableModel.SHOW_FILE;
	this.tree = null;
	this.tree_sync = tree_sync;
		  
	add = new JButton(get("MetaEditPrompt.Accumulate"));
	add.addActionListener(this);
	add.setEnabled(false);
	add.setMnemonic(KeyEvent.VK_A);
	add.setPreferredSize(BUTTON_SIZE);

	update = new JButton(get("MetaEditPrompt.Overwrite"));
	update.addActionListener(this);
	update.setEnabled(false);
	update.setMnemonic(KeyEvent.VK_P);
	update.setPreferredSize(BUTTON_SIZE);

	remove = new JButton(get("MetaEditPrompt.Remove"));
	remove.addActionListener(this);
	remove.setEnabled(false);
	remove.setMnemonic(KeyEvent.VK_R);
	remove.setPreferredSize(BUTTON_SIZE);

	expand = new JButton(get("General.Expand"));
	expand.addActionListener(this);
	expand.setEnabled(true);
	expand.setMnemonic(KeyEvent.VK_E);
	expand.setPreferredSize(new Dimension(25, 25));

	tree = new GValueTree(this, CONTROL_SIZE.width, CONTROL_SIZE.height, add, update, remove, expand);
    }
    /** Called whenever an action occurs on one of our registered buttons.
	* @param event An <strong>ActionEvent</strong> containing information about the event.
	* @see org.greenstone.gatherer.collection.CollectionManager
	* @see org.greenstone.gatherer.gui.table.GTableModel
	* @see org.greenstone.gatherer.msm.ElementWrapper
	* @see org.greenstone.gatherer.msm.MetadataSetManager
	* @see org.greenstone.gatherer.msm.MSMEvent
	* @see org.greenstone.gatherer.valuetree.GValueTree
	*/
    public void actionPerformed(ActionEvent event) {
	Object esrc = event.getSource();
	if(esrc == add) {
	    (new AppendMetadataTask()).start();
	}
	else if(esrc == update) {
	    (new UpdateMetadataTask()).start();
	}
	else if(esrc == remove) {
	    (new RemoveMetadataTask()).start();
	}
	else if(esrc == expand) {
	    EditorDialog ed = new EditorDialog();
	    String temp = ed.display(tree.getSelectedValue());
	    if(temp != null) {
		tree.setSelectedValue(temp);
	    }
	}
	//else if(esrc == assigned_metadata_view || esrc == unassigned_metadata_view) {
	//	model.changeView();
	//}
	validateControls();
    }

    private class AppendMetadataTask
	extends Thread {

	public AppendMetadataTask() { }

	public void run() {
	    // Check the new metadata is valid
	    ElementWrapper element = tree.getSelectedMetadataElement();
	    String value = tree.getSelectedValue();
	    if(records != null && element != null && value != null) {
		// Check the records, and if they are folders then display the warning.
		if(!records[0].isLeaf()) {
		    WarningDialog dialog = new WarningDialog("warning.DirectoryLevelMetadata", true);
		    if(dialog.display() == JOptionPane.OK_OPTION) {
			Gatherer.c_man.getCollection().msm.addMetadata(System.currentTimeMillis(), records, element, value);
		    }
		    dialog.dispose();
		    dialog = null;
		}
		else {
		    Gatherer.c_man.getCollection().msm.addMetadata(System.currentTimeMillis(), records, element, value);
		}
		GValueNode value_node = ((GValueModel) tree.getModel()).addValue(value);
		model.setSelectedMetadata(new Metadata(element, value_node));
	    }
	}
    }

    private class UpdateMetadataTask
	extends Thread {

	public UpdateMetadataTask() { }

	public void run() {
	    // You can only update if there is a selected_metadata and
	    // you have valid values in all fields.
	    ElementWrapper element = tree.getSelectedMetadataElement();
	    String value = tree.getSelectedValue();
	    if(selected_metadata != null && records != null && element != null && value != null) {
		selected_metadata = Gatherer.c_man.getCollection().msm.updateMetadata(System.currentTimeMillis(), selected_metadata, records, value, MetaEditPrompt.OVERWRITE, selected_metadata.isFileLevel());
	    }
	    GValueNode value_node = ((GValueModel) tree.getModel()).addValue(value);
	    model.setSelectedMetadata(new Metadata(element, value_node));
	}
    }

    private class RemoveMetadataTask
	extends Thread {

	public RemoveMetadataTask() { }

	public void run() {
	    if(selected_metadata != null && records != null) {
		Gatherer.c_man.getCollection().msm.removeMetadata(System.currentTimeMillis(), selected_metadata, records);
	    }
	}
    }

    /** Some actions can only occur after this panel has been displayed on-screen, so this method is provided to do exactly that. Such actions include the proportioning of the split panes and the setting of table column widths.
	* @see org.greenstone.gatherer.gui.table.GTableModel
	*/
    public void afterDisplay() {
	external_split.setDividerLocation(0.3);
	if(table != null) {
	    Dimension table_size = table.getPreferredSize();
	    TableColumnModel column_model = table.getColumnModel();

	    TableColumn inherited_column = column_model.getColumn(0);
	    inherited_column.setPreferredWidth(20);
	    inherited_column.setCellRenderer(new TableCellRenderer(model));

	    TableColumn element_column = column_model.getColumn(1);
	    element_column.setPreferredWidth(table_size.width / 4);
	    element_column.setCellRenderer(new TableCellRenderer(model));

	    TableColumn value_column = column_model.getColumn(2);
	    value_column.setPreferredWidth(((3 * table_size.width) / 4) - 15);
	    value_column.setCellRenderer(new TableCellRenderer(model));
	}
	table.doLayout();
	model.fireTableDataChanged();
    }
    /** Called whenever a significant change has occured in the state of the currently loaded collection.
	* @param ready <i>true</i> if there is a collection currently ready to be editing, <i>false</i> otherwise.
	* @see org.greenstone.gatherer.collection.CollectionManager
	* @see org.greenstone.gatherer.collection.CollectionModel
	* @see org.greenstone.gatherer.tree.GTree
	* @see org.greenstone.gatherer.util.TreeSynchronizer
	*/
    public void collectionChanged(boolean ready) {
	if(ready) {
	    TreeModel collection_model = Gatherer.c_man.getRecordSet();
	    args = new String[1];
	    args[0] = Gatherer.c_man.getCollection().getName();
	    collection_label.setText(get("Collection.Collection", args));
				// Update label coloring.
	    collection_label.setBackground(Gatherer.config.getColor("coloring.collection_heading_background", false));
	    collection_label.setForeground(Gatherer.config.getColor("coloring.collection_heading_foreground", false));
	    collection_tree.setModel(collection_model);
				// Update tree coloring.
	    collection_tree.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	    collection_tree.setForeground(Gatherer.config.getColor("coloring.collection_tree_foreground", false));
	    // Register our msm dependant components (or correctly their models) with msm.
	    if(model != null) {
		// remove the listener first - in case its already present
		Gatherer.c_man.getCollection().msm.removeMSMListener(model);
		Gatherer.c_man.getCollection().msm.addMSMListener(model);
	    }
	    // register the pane itself as an MSMListener, so if we add a new metadata set, it will immediately show up. - remove itself first just in case
	    Gatherer.c_man.getCollection().msm.removeMSMListener(this);
	    Gatherer.c_man.getCollection().msm.addMSMListener(this);
	}
	else {
	    collection_label.setText(get("Collection.No_Collection"));
	    collection_label.setBackground(Color.lightGray);
	    collection_label.setForeground(Color.black);
	    collection_tree.setModel(new DefaultTreeModel(new DefaultMutableTreeNode(get("Collection.No_Collection"))));
	}

	filter.setEnabled(ready);
	// Ensure that this collection tree view is synchronized with any others.
	tree_sync.add(collection_tree);
    }
    /** Used to create, connect and layout the components to be shown on this control panel.
	* @see org.greenstone.gatherer.Gatherer
	* @see org.greenstone.gatherer.collection.CollectionManager
	* @see org.greenstone.gatherer.collection.CollectionModel
	* @see org.greenstone.gatherer.file.FileOpenActionListener
	* @see org.greenstone.gatherer.gui.Filter
	* @see org.greenstone.gatherer.gui.GComboBox
	* @see org.greenstone.gatherer.gui.table.GTableModel
	* @see org.greenstone.gatherer.tree.GTree
	* @see org.greenstone.gatherer.tree.GTreeModel
	* @see org.greenstone.gatherer.valuetree.GValueTree
	*/
    public void display() {
	right_button_listener = new RightButtonListener();
	// Creation
	JPanel collection_pane = new JPanel(new BorderLayout());
	collection_pane.setMinimumSize(MINIMUM_SIZE);
	collection_pane.setPreferredSize(TREE_SIZE);

	external_split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);

	///atherer.println("\tCreating collection_label");
	args = new String[1];
	args[0] = get("Collection.No_Collection");
	collection_label = new JLabel(get("Collection.Collection"));
	collection_label.setOpaque(true);


	DragGroup group = new DragGroup();
	TreeModel collection_model = Gatherer.c_man.getRecordSet();
	if(collection_model != null) {
	    collection_tree = new DragTree("MetaEdit", collection_model, null);
	    collection_model.addTreeModelListener(this);
	}
	else {
	    collection_tree = new DragTree("MetaEdit", null);
	}
	group.add(collection_tree);
	collection_tree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
	collection_tree.putClientProperty("JTree.lineStyle", "Angled");
	collection_tree.addMouseListener(Gatherer.g_man.foa_listener);
	collection_tree.addMouseListener(right_button_listener);
	collection_tree.addTreeSelectionListener(this);
	collection_tree.addTreeExpansionListener(Gatherer.g_man.foa_listener);
	collection_tree.setLargeModel(true);
	collection_tree.setRootVisible(false);

	JScrollPane collection_scroll = new JScrollPane(collection_tree);

	filter = Gatherer.g_man.getFilter(collection_tree);
	filter.setBackground(Gatherer.config.getColor("coloring.collection_heading_background", false));
	// Change the default colours of this filters combobox.
	GComboBox fcb = filter.getComboBox();
	fcb.setBackgroundNonSelectionColor(Gatherer.config.getColor("coloring.collection_tree_background", false));
	fcb.setTextNonSelectionColor(Gatherer.config.getColor("coloring.collection_tree_foreground", false));
	fcb.setBackgroundSelectionColor(Gatherer.config.getColor("coloring.collection_selection_background", false));
	fcb.setTextSelectionColor(Gatherer.config.getColor("coloring.collection_selection_foreground", false));
	fcb = null;

	// Connection

	// Layout
	collection_pane.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(3,3,3,3), BorderFactory.createLoweredBevelBorder()));
	collection_pane.setMinimumSize(MINIMUM_SIZE);
	collection_pane.setPreferredSize(new Dimension(Gatherer.g_man.getSize().width / 3, Gatherer.g_man.getSize().height));

	// Collection Pane

	collection_pane.add(collection_label, BorderLayout.NORTH);
	collection_pane.add(collection_scroll, BorderLayout.CENTER);
	collection_pane.add(filter, BorderLayout.SOUTH);

	// Main pane
	//main_pane = new JPanel(new BorderLayout());
	//main_pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

	main_split_pane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
	main_split_pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	main_split_pane.setDividerSize(8);

	// Thumbnail - Currently disabled.
	// At least until I figure out how to render thumbnails off screen,
	// which possibly may happen with the discovery of CalpaHTML.
	/*
	  thumbnail_pane = new JPanel(new BorderLayout());

	  gatherer.debug("\tCreating thumbnail_label");
	  thumbnail_label = new JLabel(get("Thumbnail"));
	  thumbnail_label.setHorizontalAlignment(JLabel.CENTER);

	  gatherer.debug("\tCreating thumbnail_view_pane");
	  thumbnail_view_pane = new JPanel(new FlowLayout(FlowLayout.CENTER));
	  thumbnail_scroll = new JScrollPane(thumbnail_view_pane);

	  thumbnail_pane.add(thumbnail_label, BorderLayout.NORTH);
	  thumbnail_pane.add(thumbnail_scroll, BorderLayout.CENTER);
	*/

	// Metadata Table. In order to achieve what Ian envisions I'm going
	// to have to do a tricky. The table itself will sit in a stacked
	// CardLayout. Card zero will be a see-through JPanel with a message
	// smack bang in the center, something like "No Metadata". If any
	// page of the table would appear to have no metadata this card will
	// be swapped to the front.

	table_pane = new JPanel();
	table_pane.setBorder(BorderFactory.createEmptyBorder(5,0,5,0));

	JPanel table_title_pane = new JPanel();

	///atherer.println("\tCreating metadata_label");
	table_label = new JTextField(get("No_File"));
	table_label.setBackground(Color.white);
	table_label.setEditable(false);

	//activity_bar = new JProgressBar();
	//activity_bar.setPreferredSize(LABEL_SIZE);
	//activity_bar.setString(get("Ready"));
	//activity_bar.setStringPainted(true);
	//activity_bar.setValue(activity_bar.getMaximum());

	card_layout = new CardLayout();

	table_card_pane = new JPanel();

	JPanel table_pane_zero = new JPanel();
	table_pane_zero.setOpaque(false);
	
	JPanel table_pane_two = new JPanel();
	table_pane_two.setOpaque(false);

	JLabel no_file_message = new JLabel(get("No_File"));
	no_file_message.setHorizontalAlignment(JLabel.CENTER);
	no_file_message.setOpaque(false);
	no_file_message.setVerticalAlignment(JLabel.CENTER);
	
	JLabel no_metadata_message = new JLabel(get("No_Metadata"));
	no_metadata_message.setHorizontalAlignment(JLabel.CENTER);
	no_metadata_message.setOpaque(false);
	no_metadata_message.setVerticalAlignment(JLabel.CENTER);

	JPanel table_pane_one = new JPanel();

	//JPanel view_pane = new JPanel();

	//JLabel view_label = new JLabel(get("View"));
	//view_label.setPreferredSize(LABEL_SIZE);

	//JPanel view_button_pane = new JPanel();

	//assigned_metadata_view = new JToggleButton(get("View_Assigned"), Utility.OFF_ICON);
	//assigned_metadata_view.setSelectedIcon(Utility.ON_ICON);
	//assigned_metadata_view.setSelected(true);
	//assigned_metadata_view.addActionListener(this);
	//unassigned_metadata_view = new JToggleButton(get("View_Unassigned"), Utility.OFF_ICON);
	//unassigned_metadata_view.setSelectedIcon(Utility.ON_ICON);
	//unassigned_metadata_view.setSelected(true);
	//unassigned_metadata_view.addActionListener(this);
	//XORToggleButtonGroup view_group = new XORToggleButtonGroup();
	//view_group.add(assigned_metadata_view);
	//view_group.add(unassigned_metadata_view);

	///atherer.println("\tCreating metadata_table");
	table = new JTable();
	//model = new GTableModel(table, assigned_metadata_view, unassigned_metadata_view, activity_bar);
	model = new GTableModel(table);
	table.setModel(model);
	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	table.getSelectionModel().addListSelectionListener(this);
	table.setColumnSelectionAllowed(false);

	// The default JTable doesn't quite have the desired properties - when a
	// table cell is clicked on, the table is scrolled so the cell is as
	// visible as possible. This means that the left-most cells can become
	// hidden. To fix this, the MouseInputHandler in BasicTableUI is removed
	// as a MouseListener on the table, and a very slightly altered copy is
	// added in its place (TableMouseInputHandler).
	MouseListener[] mls = table.getMouseListeners();
	for (int i = 0; i < mls.length; i++) {
	    // Remove the instances of MouseInputHandler
	    if (mls[i].toString().startsWith("javax.swing.plaf.basic.BasicTableUI$MouseInputHandler@")) {
		table.removeMouseListener(mls[i]);
	    }
	}
	// Add the slightly-altered mouse listener
	table.addMouseListener(new TableMouseInputHandler());

	JScrollPane table_scroll = new JScrollPane(table);
	table_scroll.getViewport().setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	table_scroll.setOpaque(true);

	// Control pane
	///atherer.println("\tCreating Metadata Controls");

	control_pane = new JPanel();
	control_pane.setBorder(BorderFactory.createLoweredBevelBorder());
	control_pane.setMinimumSize(MINIMUM_SIZE);
	control_pane.setSize(CONTROL_SIZE);
	control_pane.setPreferredSize(CONTROL_SIZE);

	card_layout2 = new CardLayout();

	JPanel tools_off_pane = new JPanel();

	JLabel tools_off_label = new JLabel(get("No_Metadata_Element"));
	tools_off_label.setHorizontalAlignment(JLabel.CENTER);
	tools_off_label.setOpaque(false);
	tools_off_label.setVerticalAlignment(JLabel.CENTER);

	JPanel tools_on_pane = new JPanel();
	tools_on_pane.setMinimumSize(MINIMUM_SIZE);
	tools_on_pane.setSize(TABLE_SIZE);
	tools_on_pane.setPreferredSize(TABLE_SIZE);

	// Layout.

	// Metaedit controls
	tools_off_pane.setLayout(new BorderLayout());
	tools_off_pane.add(tools_off_label, BorderLayout.CENTER);

	tools_on_pane.setLayout(new BorderLayout());
	tools_on_pane.add(tree, BorderLayout.CENTER);

	control_pane.setLayout(card_layout2);
	control_pane.add(tools_off_pane, TOOLS_OFF);
	control_pane.add(tools_on_pane, TOOLS_ON);

	// Table components
	table_title_pane.setBorder(BorderFactory.createEmptyBorder(0,0,5,0));
	table_title_pane.setLayout(new BorderLayout());
	//table_title_pane.setLayout(new GridLayout(1,2,5,0));
	table_title_pane.add(table_label, BorderLayout.CENTER);
	//table_title_pane.add(activity_bar);

	table_pane_zero.setLayout(new BorderLayout());
	table_pane_zero.add(no_file_message, BorderLayout.CENTER);

	table_pane_one.setLayout(new BorderLayout());
	table_pane_one.add(table_scroll, BorderLayout.CENTER);

	table_pane_two.setLayout(new BorderLayout());
	table_pane_two.add(no_metadata_message, BorderLayout.CENTER);
	
	table_card_pane.setLayout(card_layout);
	table_card_pane.add(table_pane_zero, CARD_ZERO);
	table_card_pane.add(table_pane_one, CARD_ONE);
	table_card_pane.add(table_pane_two, CARD_TWO);
	//view_button_pane.setLayout(new GridLayout(1,2,5,0));
	//view_button_pane.add(assigned_metadata_view);
	//view_button_pane.add(unassigned_metadata_view);

	//view_pane.setBorder(BorderFactory.createEmptyBorder(5,0,0,0));
	//view_pane.setLayout(new BorderLayout());
	//view_pane.add(view_label, BorderLayout.WEST);
	//view_pane.add(view_button_pane, BorderLayout.CENTER);

	table_pane.setLayout(new BorderLayout());
	table_pane.add(table_title_pane, BorderLayout.NORTH);
	table_pane.add(table_card_pane, BorderLayout.CENTER);
	//table_pane.add(view_pane, BorderLayout.SOUTH);

	//main_split_pane.add(control_pane, JSplitPane.TOP);
	//main_split_pane.add(table_pane, JSplitPane.BOTTOM);
	main_split_pane.add(table_pane, JSplitPane.TOP);
	main_split_pane.add(control_pane, JSplitPane.BOTTOM);
	main_split_pane.setDividerLocation(250);

	external_split.add(collection_pane, JSplitPane.LEFT);//BorderLayout.WEST);
	//this.add(main_pane, BorderLayout.CENTER);
	external_split.add(main_split_pane, JSplitPane.RIGHT);//BorderLayout.CENTER);

	this.setLayout(new BorderLayout());
	this.add(external_split, BorderLayout.CENTER);
    }

    /** Method that is called whenever an element within a set is changed or modified. - needed for MSMListener
     * @param event A <strong>MSMEvent</strong> containing details of the event that caused this message to be fired.
     */
    public void elementChanged(MSMEvent event) {}

    /** Ensures a certain tree path is expanded, visible and selected within the collection tree.
	* @param path The <strong>TreePath</strong> to make visible.
	* @see org.greenstone.gatherer.tree.GTree
	*/
    public void expandPath(TreePath path) {
	collection_tree.expandPath(path);
	collection_tree.setSelectionPath(path);
    }
    /** Called to inform this control panel that it has just gained focus as an effect of the user clicking on its tab.
	* @see org.greenstone.gatherer.tree.GTree
	*/
    public void gainFocus() {
	// Use this opportunity to update the table model etc.
	valueChanged(new TreeSelectionEvent(this, null, false, null, null));
	// tree.setElementModel(Gatherer.c_man.getCollection().msm.getElementModel());
	// Update the menubars idea of whats been selected
	if(collection_tree != null) {
	    if(collection_tree.isSelectionEmpty()) {
		Gatherer.g_man.menu_bar.setMetaAuditSuffix(null);
	    }
	    else {
		Gatherer.g_man.menu_bar.setMetaAuditSuffix(collection_tree.getSelectionDetails());
	    }
	}
	// Update the meta-audit view to show the current selection, if any.
	Gatherer.g_man.meta_audit.setRecords(records);
    }
    /** Used to determine if this class contains a reference to an existing file record that matches the presumtively new record, and if one is found, returns it. Note that there is no need to search the metadata table model, as any file record that exists there must also exist in the records selection.
	 * @param new_record The <strong>FileNode</strong> which may or may not be the first occurance of a certain record.
	 * @return The matching <strong>FileNode</strong> if there is one, or <i>null</i>.
	 */
    public FileNode getAnyExistingRecord(FileNode new_record) {
	FileNode match = null;
	for(int i = 0; records != null && i < records.length && match == null; i++) {
	    if(records[i].equals(new_record)) {
		match = records[i];
	    }
	}
	return match;
    }

    /** Retrieve the currently selected records.
	* @return A <strong>FileNode[]</strong> containing the current, and up to date, selection.
	*/
    public FileNode[] getSelectedNode() {
	return records;
    }

    /** Retrieve the currently selected metadata.
	* @return The <strong>Metadata</strong> in question.
	*/
    public Metadata getSelectedMetadata() {
	return selected_metadata;
    }

    /** Called whenever the metadata value changes in some way, such as the addition of a new value. - needed for MSMListener
     * @param event A <strong>MSMEvent</strong> containing details of the event that caused this message to be fired.
     */
    public void metadataChanged(MSMEvent event) {}

    /** Method that is called whenever the metadata set collection changes in some way, such as the addition of a new set or the merging of two sets. - needed for MSMListener
     * @param event A <strong>MSMEvent</strong> containing details of the event that caused this message to be fired.
     */
    public void setChanged(MSMEvent event){
	// call the collection tree value changed - we need to do the same stuff here
	valueChanged((TreeSelectionEvent) null);

    }

    public void setSelection(File file) {
	collection_tree.setSelection(file);
    }

    /** Determine the bounds of the selected metadata within the table, used during search and replace actions.
	* @param metadata The <strong>Metadata</strong> to search for.
	*/
    public Rectangle setSelectedMetadata(Metadata metadata) {
	return model.setSelectedMetadata(metadata);
    }

    public void refreshTrees() {
	collection_tree.refresh(null); // Refresh entire tree.
    }

    /* Called when nodes have been modified, but the model hasn't changed.
	* @param event Everything you ever wanted to know about model changes in one tidy <strong>TreeModelEvent</strong> package.
	*/
    public void treeNodesChanged(TreeModelEvent event) {
	has_changed = true;
    }

    /* Called when nodes have been added to the model.
	  * @param event Everything you ever wanted to know about model changes in one tidy <strong>TreeModelEvent</strong> package.
	  */
    public void treeNodesInserted(TreeModelEvent event) {
	has_changed = true;
    }

    /* Called when nodes have been removed from the model.
	  * @param event Everything you ever wanted to know about model changes in one tidy <strong>TreeModelEvent</strong> package.
	  */
    public void treeNodesRemoved(TreeModelEvent event) {
	has_changed = true;
    }

    /** Called when nodes have been rearranged within the model.
	  * @param event Everything you ever wanted to know about model changes in one tidy <strong>TreeModelEvent</strong> package.
	  */
    public void treeStructureChanged(TreeModelEvent event) {
	has_changed = true;
    }
    /** Called to ensure that if the tree model has changed in some way, the view has been updated.
	* @see org.greenstone.gatherer.tree.GTree
	*/
    public void updateTree() {
	if(has_changed) {
	    tree.updateUI();
	    has_changed = false;
	}
    }

    /** This method enables or diables the various controls based upon the state of the system. This method should be called after every change which could affect the GUI.
	  */
    public void validateControls() {
	validateControls(true);
    }
    public void validateControls(boolean and_tree) {
	validateMetadataTable();
	// Validate card_layout_2
	if (selected_metadata != null) {
	    card_layout2.show(control_pane, TOOLS_ON);
	}
	else {
	    card_layout2.show(control_pane, TOOLS_OFF);
	}
	if(and_tree) {
	    // Validate the buttons in the lower pane
	    if (selected_metadata == null) {
		// Nothing selected
		return;
	    }

	    // Does the metadata element have no current value?
	    GValueNode value_node = selected_metadata.getValueNode();
	    if (value_node == null) {
		// Can only append if something has been entered
		add.setEnabled((tree.getSelectedValue().length() > 0));
		// Nothing to replace or remove
		update.setEnabled(false);
		remove.setEnabled(false);
	    }

	    else {
		// Check if the text in the value field is the same as the metadata value
		if (tree.getSelectedValue().equals(value_node.getFullPath())) {
		    // Can't append or replace
		    add.setEnabled(false);
		    update.setEnabled(false);
		}
		else {
		    // Can append or replace
		    add.setEnabled(true);
		    update.setEnabled(true);
		}

		// Can only remove if the metadata is file level
		remove.setEnabled(selected_metadata.isFileLevel());
	    }
	}
    }

    /** Validate the metadata table area, and if no metadata is available display the no metadata panel. */
    public void validateMetadataTable() {
	// Validate card_layout
	if (records == null) {
	    card_layout.show(table_card_pane, CARD_ZERO);
	} else if (model.getRowCount() == 0) {
	    card_layout.show(table_card_pane, CARD_TWO);
	} else {
	    card_layout.show(table_card_pane, CARD_ONE);
	}
	/*
	if(records != null && model.getRowCount() > 0) {
	    card_layout.show(table_card_pane, CARD_ONE);
	}
	else {
	    card_layout.show(table_card_pane, CARD_ZERO);
	    }*/
    }

    /** Another in a long list of listeners, this one is called whenever the
     * selection within the JTable changes. When it does we load the new data
     * and set the selected_metadata if applicable (ie the metadata is non-
     * empty).
     * @param event The <strong>ListSelectionEvent</strong> which contains all the information about the event that has been fired.
     * @see org.greenstone.gatherer.gui.table.GTableModel
     * @see org.greenstone.gatherer.valuetree.GValueTree
     */
    public void valueChanged(ListSelectionEvent event) {
	// We only want to handle one event per selection, so wait for the value to stabilise
	ListSelectionModel lsm = table.getSelectionModel();
	if (lsm.getValueIsAdjusting() == false) {
	    // We have a SINGLE_SELECTION model set so there is at most one
	    // selection. Get the offending row.
	    int selected = table.getSelectedRow();
	    selected_metadata = model.getMetadataAtRow(selected);
	    if(selected_metadata != null) { // Check something is selected.
		tree.setSelectedMetadataElement(selected_metadata.getElement());
		GValueNode value_node = selected_metadata.getValueNode();
		if(value_node != null) {
		    tree.setSelectedValue(value_node.getFullPath());
		}
		else {
		    tree.setSelectedValue("");
		}
	    }
	    validateControls(true);
	}
    }

    /** Called whenever the value tree of an metadata element changes in some way, such as the addition of a new value. - needed for MSMListener
     * @param event A <strong>MSMEvent</strong> containing details of the event that caused this message to be fired.
     */
    public void valueChanged(MSMEvent event) {}

    /** This method is called whenever the selection within the collection
     * tree changes. This causes the table to be rebuilt with a new model
     * @param event A <strong>TreeSelectionEvent</strong> containing information about the event.
     * @see org.greenstone.gatherer.Gatherer
     * @see org.greenstone.gatherer.collection.FileNode
     * @see org.greenstone.gatherer.gui.GUIManager
     * @see org.greenstone.gatherer.gui.MenuBar
     * @see org.greenstone.gatherer.gui.table.GTableModel
     * @see org.greenstone.gatherer.msm.MetadataSetManager
     * @see org.greenstone.gatherer.tree.GTree
     * @see org.greenstone.gatherer.util.ArrayTools
     */
    public void valueChanged(TreeSelectionEvent event) {
	// System.err.println("\n(MEP) valueChanged(TreeSelectionEvent)...");

	// Don't bother if this isn't the selected view
	if (Gatherer.g_man.getSelectedView() != this) {
	    return;
	}

	if(collection_tree.getSelectionCount() > 0) {
	    records = null;
	    TreePath paths[] = collection_tree.getSelectionPaths();
	    for(int i = 0; i < paths.length; i++) {
		FileNode record = (FileNode) paths[i].getLastPathComponent();
		records = ArrayTools.add(records, record);
	    }

	    table_label.setText(collection_tree.getSelectionDetails());

	    // Remove old model from msm
	    Gatherer.c_man.getCollection().msm.removeMSMListener(model);
	    // Create new model
	    model = new GTableModel(table, records);
	    table.setModel(model);
	}
	else {
	    records = null;
	    table_label.setText(get("No_File"));
	    // Remove old model from msm
	    if(Gatherer.c_man.ready()) {
		Gatherer.c_man.getCollection().msm.removeMSMListener(model);
	    }
	    // Create new model
	    model = new GTableModel(table);
	    table.setModel(model);
	}

	table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
	if(table != null) {
	    Dimension table_size = table.getPreferredSize();
	    TableColumnModel column_model = table.getColumnModel();

	    TableColumn inherited_column = column_model.getColumn(0);
	    inherited_column.setPreferredWidth(20);
	    inherited_column.setCellRenderer(new TableCellRenderer(model));
			 
	    TableColumn element_column = column_model.getColumn(1);
	    element_column.setPreferredWidth(TABLE_SIZE.width / 4);
	    element_column.setCellRenderer(new TableCellRenderer(model));
			 
	    TableColumn value_column = column_model.getColumn(2);
	    value_column.setPreferredWidth(((3 * TABLE_SIZE.width) / 4) - 15);
	    value_column.setCellRenderer(new TableCellRenderer(model));
	}
	table.doLayout();
	validateControls();
    }


    /** Retrieve a phrase from the Dictionary based on the given key.
	* @param key The <strong>String</strong> which maps to the phrase to retrieve.
	* @return The desired phrase as a <strong>String</strong>, or possibly an error message if no such phrase exists.
	*/
    private String get(String key) {
	return get(key, null);
    }
    /** Retrieve a phrase from the Dictionary based on the given key and filled in with the given arguments.
	* @param key The <strong>String</strong> which maps to the phrase to retrieve.
	* @param args A <strong>String[]</strong> of arguments to be inserted in the phrase.
	* @return The desired phrase as a <strong>String</strong>, or possibly an error message if no such phrase exists.
	* @see org.greenstone.gatherer.Dictionary
	* @see org.greenstone.gatherer.Gatherer
	*/
    private String get(String key, String args[]) {
	if(key.indexOf('.') == -1) {
	    key = "MetaEdit." + key;
	}
	return Gatherer.dictionary.get(key,args);
    }


    /** A direct copy of javax.swing.plaf.basic.BasicTableUI$MouseInputHandler, except for one change in adjustFocusAndSelection(). The purpose of this change is to always keep the left-most cells of the table row selected visible. */
    private class TableMouseInputHandler implements MouseInputListener
    {
        // Component receiving mouse events during editing.
        // May not be editorComponent.
        private Component dispatchComponent;
	private boolean selectedOnPress;

	//  The Table's mouse listener methods.

        public void mouseClicked(MouseEvent e) {}

        private void setDispatchComponent(MouseEvent e) {
            Component editorComponent = table.getEditorComponent();
            Point p = e.getPoint();
            Point p2 = SwingUtilities.convertPoint(table, p, editorComponent);
            dispatchComponent = SwingUtilities.getDeepestComponentAt(editorComponent,
								     p2.x, p2.y);
        }

        private boolean repostEvent(MouseEvent e) {
	    // Check for isEditing() in case another event has
	    // caused the editor to be removed. See bug #4306499.
            if (dispatchComponent == null || !table.isEditing()) {
                return false;
            }
            MouseEvent e2 = SwingUtilities.convertMouseEvent(table, e, dispatchComponent);
            dispatchComponent.dispatchEvent(e2);
            return true;
        }

        private void setValueIsAdjusting(boolean flag) {
            table.getSelectionModel().setValueIsAdjusting(flag);
            table.getColumnModel().getSelectionModel().setValueIsAdjusting(flag);
        }

	private boolean shouldIgnore(MouseEvent e) {
	    return e.isConsumed() || (!(SwingUtilities.isLeftMouseButton(e) && table.isEnabled()));
	}

        public void mousePressed(MouseEvent e) {
	    if (e.isConsumed()) {
		selectedOnPress = false;
		return;
	    }
	    selectedOnPress = true;
	    adjustFocusAndSelection(e);
	}

	void adjustFocusAndSelection(MouseEvent e) {
	    if (shouldIgnore(e)) {
	        return;
	    }

            Point p = e.getPoint();
            int row = table.rowAtPoint(p);
            int column = table.columnAtPoint(p);
	    // The autoscroller can generate drag events outside the Table's range.
            if ((column == -1) || (row == -1)) {
                return;
            }

            if (table.editCellAt(row, column, e)) {
                setDispatchComponent(e);
                repostEvent(e);
            }
	    else if (table.isRequestFocusEnabled()) {
                table.requestFocus();
	    }

            CellEditor editor = table.getCellEditor();
            if (editor == null || editor.shouldSelectCell(e)) {
		boolean adjusting = (e.getID() == MouseEvent.MOUSE_PRESSED) ? true : false;
                setValueIsAdjusting(adjusting);
		if (column == 0) {
		    table.changeSelection(row, 0, e.isControlDown(), e.isShiftDown());
		}
		else {
		    table.changeSelection(row, 1, e.isControlDown(), e.isShiftDown());
		}
                // table.changeSelection(row, column, e.isControlDown(), e.isShiftDown());
	    }
        }

        public void mouseReleased(MouseEvent e) {
	    if (selectedOnPress) {
		if (shouldIgnore(e)) {
		    return;
		}

		repostEvent(e);
		dispatchComponent = null;
		setValueIsAdjusting(false);
	    } else {
		adjustFocusAndSelection(e);
	    }
        }


        public void mouseEntered(MouseEvent e) {}

        public void mouseExited(MouseEvent e) {}

	//  The Table's mouse motion listener methods.

        public void mouseMoved(MouseEvent e) {}

        public void mouseDragged(MouseEvent e) {
	    if (shouldIgnore(e)) {
	        return;
	    }

            repostEvent(e);

            CellEditor editor = table.getCellEditor();
            if (editor == null || editor.shouldSelectCell(e)) {
                Point p = e.getPoint();
                int row = table.rowAtPoint(p);
                int column = table.columnAtPoint(p);
	        // The autoscroller can generate drag events outside the Table's range.
                if ((column == -1) || (row == -1)) {
                    return;
                }
	        table.changeSelection(row, column, false, true);
            }
        }
    }


    /** A listener for right mouse button clicks over the collection tree. */
    private class RightButtonListener
	extends MouseAdapter {
	/** Called whenever a mouse click occurs (right or left) over a target component.
	 * @param event A <strong>MouseEvent</strong> containing further information about the mouse click action.
	 * @see org.greenstone.gatherer.gui.MetaEditPane.RightButtonMenu
	 */
	public void mouseClicked(MouseEvent event) {
	    if(SwingUtilities.isRightMouseButton(event)) {
		new RightButtonMenu(event);
	    }
	}
    }
    /** Provides a popup menu to display when a right mouse button click is detected over the collection tree. */
    private class RightButtonMenu
	extends JPopupMenu
	implements ActionListener {
	/** Constructor.
	 * @param event The <strong>MouseEvent</strong> that triggered the creation of this menu. Used to determine where the menu will be located.
		*/
	public RightButtonMenu(MouseEvent event) {
	    super();
				// Creation
	    JMenuItem show_metaaudit = new JMenuItem(get("Menu.Metadata_View") + " " + collection_tree.getSelectionDetails(), KeyEvent.VK_V);
	    show_metaaudit.addActionListener(this);
	    add(show_metaaudit);
				// Display
	    show(collection_tree, event.getX(), event.getY());
	    show_metaaudit = null;
	    args = null;

	}
	/** Called whenever a user clicks on the single menu item, view all assigned metadata.
		* @param event An <strong>ActionEvent</strong> containing further information about the action.
		* @see org.greenstone.gatherer.Gatherer
		* @see org.greenstone.gatherer.gui.GUIManager
		*/
	public void actionPerformed(ActionEvent event) {
	    Gatherer.g_man.showMetaAuditBox();
	}
    }

}
